/*
	version.h (12.06.10)
	Version constants.

	Copyright (C) 2009, 2010  Andrew Nayenko

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef VERSION_H_INCLUDED
#define VERSION_H_INCLUDED

#define EXFAT_VERSION_MAJOR 0
#define EXFAT_VERSION_MINOR 9
#define EXFAT_VERSION_PATCH 7

#endif /* ifndef VERSION_H_INCLUDED */
